/*
 * Copyright (c) 2010 Yoshikazu Kuramochi
 * All rights reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package ch.kuramo.javie.effects.noiseGrain;

import javax.media.opengl.GL2;
import javax.media.opengl.glu.GLU;
import javax.vecmath.Matrix3d;

import ch.kuramo.javie.api.IAnimatableBoolean;
import ch.kuramo.javie.api.IAnimatableDouble;
import ch.kuramo.javie.api.IAnimatableEnum;
import ch.kuramo.javie.api.IAnimatableInteger;
import ch.kuramo.javie.api.IAnimatableVec2d;
import ch.kuramo.javie.api.IShaderProgram;
import ch.kuramo.javie.api.IVideoBuffer;
import ch.kuramo.javie.api.RenderResolution;
import ch.kuramo.javie.api.Vec2d;
import ch.kuramo.javie.api.VideoBounds;
import ch.kuramo.javie.api.annotations.Effect;
import ch.kuramo.javie.api.annotations.Property;
import ch.kuramo.javie.api.annotations.ShaderSource;
import ch.kuramo.javie.api.services.IShaderRegistry;
import ch.kuramo.javie.api.services.IVideoEffectContext;
import ch.kuramo.javie.effects.VideoEffectUtil;

import com.google.inject.Inject;

@Effect(id="ch.kuramo.javie.FractalNoise", category=Effect.NOISE_AND_GRAIN)
public class FractalNoise {

	@ShaderSource(program=false)
	public static final String noise = "noise.frag";

	@ShaderSource(attach="noise")
	public static final String[] FRACTAL_NOISE = {
		"uniform int fractalType;",
		"uniform float contrast;",
		"uniform float brightness;",
		"uniform int overflow;",
		"uniform int complexity;",
		"uniform float influences[20];",
		"uniform mat3 matrices[20];",
		"uniform float evolution;",
		"uniform float cycle;",
		"",
		"float noise(vec3 P);",
		"",
		"void main(void)",
		"{",
		"	float sum = 0.0;",
		"	vec3 pt = vec3(gl_FragCoord.st/128.0, 1.0);",
		"",
		"	if (fractalType == 0) {",
		"		if (cycle == 0.0) {",
		"			for (int i = 0; i < complexity; ++i) {",
		"				sum += influences[i]*noise(vec3((matrices[i]*pt).xy, evolution));",
		"			}",
		"		} else {",
		"			for (int i = 0; i < complexity; ++i) {",
		"				sum += influences[i]/cycle*(",
		"						(cycle-evolution)*noise(vec3((matrices[i]*pt).xy, evolution))",
		"					  + evolution*noise(vec3((matrices[i]*pt).xy, evolution-cycle)));",
		"			}",
		"		}",
		"		sum *= 2.0;",
		"	} else if (fractalType == 1) {",
		"		if (cycle == 0.0) {",
		"			for (int i = 0; i < complexity; ++i) {",
		"				sum += influences[i]*abs(noise(vec3((matrices[i]*pt).xy, evolution)));",
		"			}",
		"		} else {",
		"			for (int i = 0; i < complexity; ++i) {",
		"				sum += influences[i]/cycle*(",
		"						(cycle-evolution)*abs(noise(vec3((matrices[i]*pt).xy, evolution)))",
		"					  + evolution*abs(noise(vec3((matrices[i]*pt).xy, evolution-cycle))));",
		"			}",
		"		}",
		"		sum = sum * 2.0 - 1.0;",
		"	}",
		"",
		"	float t = sum * contrast + brightness;",
		"",
		"	if (overflow == 0) {",									// CLIP
		"		t = min(max(0.5+0.5*t, 0.0), 1.0);",
		"	} else if (overflow == 1) {",							// SOFT_CLAMP
		"		t = 0.5+0.3183099*atan(t);",
		"	} else if (overflow == 2) {",							// WRAP_BACK
		"		t = abs(0.5+0.5*t);",
		"		t = 1.0 - abs(t - 2.0*float(int(t*0.5)) - 1.0);",
		"	} else {",												// HDR
		"		t = 0.5+0.5*t;",
		"	}",
		"	",
		"	gl_FragColor = vec4(t, t, t, 1.0);",
		"}",
	};


	public enum FractalType { BASIC, TURBULENT }

	public enum Overflow { CLIP, SOFT_CLAMP, WRAP_BACK, HDR }

	public enum BlendMode {
		NONE,
		NORMAL,
		//DISSOLVE,
		//DANCING_DISSOLVE,

		DARKEN,
		MULTIPLY,
		COLOR_BURN,
		LINEAR_BURN,
		DARKER_COLOR,

		ADD,
		LIGHTEN,
		SCREEN,
		COLOR_DODGE,
		LINEAR_DODGE,
		LIGHTER_COLOR,

		OVERLAY,
		SOFT_LIGHT,
		HARD_LIGHT,
		LINEAR_LIGHT,
		VIVID_LIGHT,
		PIN_LIGHT,
		HARD_MIX,

		DIFFERENCE,
		EXCLUSION,

		HUE,
		SATURATION,
		//COLOR,
		LUMINOSITY,

		//STENCIL_ALPHA,
		STENCIL_LUMA,
		//SILHOUETTE_ALPHA,
		SILHOUETTE_LUMA;
	}


	@Property
	private IAnimatableEnum<FractalType> fractalType;

//	@Property
//	private IAnimatableEnum<NoiseType> noiseType;

	@Property
	private IAnimatableBoolean invert;

	@Property(value="100", min="0", max="10000")
	private IAnimatableDouble contrast;

	@Property(value="0", min="-10000", max="10000")
	private IAnimatableDouble brightness;

	@Property("HDR")
	private IAnimatableEnum<Overflow> overflow;

	@Property
	private IAnimatableDouble rotation;

	@Property(value="100", min="1", max="10000")
	private IAnimatableVec2d scale;

	@Property
	private IAnimatableVec2d offset;

	@Property(value="6", min="1", max="20")
	private IAnimatableDouble complexity;

	@Property(value="70", min="0", max="10000")
	private IAnimatableDouble subInfluence;

	@Property(value="56", min="10", max="10000")
	private IAnimatableDouble subScale;

	@Property
	private IAnimatableDouble subRotation;

	@Property
	private IAnimatableVec2d subOffset;

	@Property
	private IAnimatableBoolean centerSubscale;

	@Property
	private IAnimatableDouble evolution;

	@Property(value="0", min="0", max="100")
	private IAnimatableInteger cycle;

	@Property(value="0", min="0", max="100000")
	private IAnimatableInteger randomSeed;

	@Property(value="100", min="0", max="100")
	private IAnimatableDouble opacity;

	@Property("NORMAL")
	private IAnimatableEnum<BlendMode> blendMode;


	private final IVideoEffectContext context;

	private final IShaderRegistry shaders;

	private final IShaderProgram program;

	@Inject
	public FractalNoise(IVideoEffectContext context, IShaderRegistry shaders) {
		this.context = context;
		this.shaders = shaders;
		program = shaders.getProgram(FractalNoise.class, "FRACTAL_NOISE");
	}

	public IVideoBuffer doVideoEffect() {
		GL2 gl = context.getGL().getGL2();
		GLU glu = context.getGLU();

		IVideoBuffer input = null;
		IVideoBuffer buf1 = null;
		IVideoBuffer buf2 = null;
		try {
			input = context.doPreviousEffect();
			if (input.getBounds().isEmpty()) {
				IVideoBuffer result = input;
				input = null;
				return result;
			}

			buf1 = doFractalNoise(gl, glu, input);

			BlendMode blendMode = context.value(this.blendMode);
			double opacity = context.value(this.opacity) / 100;
			if ((blendMode == BlendMode.NONE || blendMode == BlendMode.NORMAL) && opacity == 1.0) {
				buf2 = buf1;
				buf1 = null;
			} else {
				buf2 = doBlend(gl, glu, buf1, input, blendMode, opacity);
			}

			IVideoBuffer result = buf2;
			buf2 = null;
			return result;

		} finally {
			if (input != null) input.dispose();
			if (buf1 != null) buf1.dispose();
			if (buf2 != null) buf2.dispose();
		}
	}

	private IVideoBuffer doFractalNoise(final GL2 gl, GLU glu, IVideoBuffer input) {
		int permTexture = FractalNoiseTexture.getPermTexture(gl);

		VideoBounds bounds = input.getBounds();
		final int w = bounds.width;
		final int h = bounds.height;

		RenderResolution resolution = context.getRenderResolution();


		final FractalType fractalType = context.value(this.fractalType);
		boolean invert = context.value(this.invert);
		double contrast = context.value(this.contrast);
		double brightness = context.value(this.brightness);
		final Overflow overflow = context.value(this.overflow);
		double rotation = Math.toRadians(context.value(this.rotation));
		Vec2d scale = resolution.scale(context.value(this.scale));
		Vec2d offset = resolution.scale(context.value(this.offset));
		double complexity = context.value(this.complexity);
		double subInfluence = context.value(this.subInfluence);
		double subScale = context.value(this.subScale);
		double subRotation = Math.toRadians(context.value(this.subRotation));
		Vec2d subOffset = context.value(this.subOffset);
		boolean centerSubscale = context.value(this.centerSubscale);
		double evolution = context.value(this.evolution);
		int cycle = context.value(this.cycle);
		int randomSeed = context.value(this.randomSeed);


		offset = new Vec2d(offset.x - bounds.x, offset.y - bounds.y);

		final float fcontrast = (float)((invert ? -1 : 1) * contrast / 100);
		final float fbrightness = (float)(brightness / 100);
		subInfluence /= 100;

		evolution += randomSeed*367;
		evolution /= 180;
		if (cycle != 0) {
			cycle *= 2;
			evolution %= cycle;
		}

		final float fevolution = (float) evolution;
		final int fcycle = cycle;

		double cosRot = Math.cos(rotation);
		double sinRot = Math.sin(rotation);
		Matrix3d transformMat = new Matrix3d(100/scale.x, 0.0, 0.0, 0.0, 100/scale.y, 0.0, 0.0, 0.0, 1.0);	// scale
		transformMat.mul(new Matrix3d(cosRot, sinRot, 0.0, -sinRot, cosRot, 0.0, 0.0, 0.0, 1.0));			// rotation
		transformMat.mul(new Matrix3d(1.0, 0.0, -offset.x/128, 0.0, 1.0, -offset.y/128, 0.0, 0.0, 1.0));	// offset

		double cosSubRot = Math.cos(subRotation);
		double sinSubRot = Math.sin(subRotation);
		Matrix3d subScaleMat1 = new Matrix3d(100/subScale, 0.0, 0.0, 0.0, 100/subScale, 0.0, 0.0, 0.0, 1.0);			// subScale  1回分
		Matrix3d subRotMat1 = new Matrix3d(cosSubRot, sinSubRot, 0.0, -sinSubRot, cosSubRot, 0.0, 0.0, 0.0, 1.0);		// subRotation  1回分
		Matrix3d subOffsetMat1 = new Matrix3d(1.0, 0.0, -subOffset.x/128, 0.0, 1.0, -subOffset.y/128, 0.0, 0.0, 1.0);	// subOffset  1回分

		Matrix3d subScaleMat = new Matrix3d();		// subScaleMat1 の累積
		subScaleMat.setIdentity();
		Matrix3d subRotMat = new Matrix3d();		// subRotMat1 の累積
		subRotMat.setIdentity();
		Matrix3d subOffsetMat = new Matrix3d();		// subOffsetMat1 の累積
		if (centerSubscale) {
			subOffsetMat.set(subOffsetMat1);
		} else {
			subOffsetMat.setIdentity();
		}

		final int ceiledComplexity = (int)Math.ceil(complexity); 
		final float[] influences = new float[ceiledComplexity];
		float infTotal = 0;
		final float[] matrices = new float[9*ceiledComplexity];

		for (int i = 0; i < ceiledComplexity; ++i) {
			influences[i] = (float)Math.pow(subInfluence, i);
			if (i+1 > complexity) {
				influences[i] *= complexity-i;
			}

			// 複雑度をアニメーションさせたとき、より滑らかに変化する計算式の例
			//double t = complexity - i;
			//influences[i] = (float)(Math.pow(subInfluence, i)
			//		*(Math.atan(Math.PI/2*(t-1))+Math.atan(Math.PI/2))/(Math.PI/2+Math.atan(Math.PI/2)));

			infTotal += influences[i];

			Matrix3d mat = new Matrix3d(subOffsetMat);
			mat.mul(subScaleMat);
			mat.mul(subRotMat);
			mat.mul(transformMat);

			for (int j = 0; j < 9; ++j) {
				matrices[i*9+j] = (float)mat.getElement(j%3, j/3);
												// (j/3, j%3) ではないのは transpose しながら取得しているため。
			}

			subScaleMat.mul(subScaleMat1);
			subRotMat.mul(subRotMat1);

			if (!centerSubscale) {
				subOffsetMat.mul(new Matrix3d(
						1.0, 0.0, w/resolution.scale/128,	// TODO ここの値はこれでいいのか？
						0.0, 1.0, h/resolution.scale/128,	//      あと、128で割らないとなぜか1/3解像度だけおかしくなる。
						0.0, 0.0, 1.0));
				subOffsetMat.mul(subOffsetMat1);
			}
		}
		for (int i = 0; i < ceiledComplexity; ++i) {
			influences[i] /= infTotal;
		}


		VideoEffectUtil.ortho2D(gl, glu, w, h);

		IVideoBuffer buffer = context.createVideoBuffer(bounds);
		try {
			gl.glFramebufferTexture2D(GL2.GL_FRAMEBUFFER,
					GL2.GL_COLOR_ATTACHMENT0, GL2.GL_TEXTURE_RECTANGLE, buffer.getTexture(), 0);
			gl.glDrawBuffer(GL2.GL_COLOR_ATTACHMENT0);

			gl.glActiveTexture(GL2.GL_TEXTURE0);
			gl.glBindTexture(GL2.GL_TEXTURE_2D, permTexture);
//			gl.glActiveTexture(GL2.GL_TEXTURE1);
//			gl.glBindTexture(GL2.GL_TEXTURE_2D, gradTexture);

			program.useProgram(new Runnable() {
				public void run() {
					gl.glUniform1i(program.getUniformLocation("permTexture"), 0);
//					gl.glUniform1i(program.getUniformLocation("gradTexture"), 1);
					gl.glUniform1i(program.getUniformLocation("fractalType"), fractalType.ordinal());
					gl.glUniform1f(program.getUniformLocation("contrast"), fcontrast);
					gl.glUniform1f(program.getUniformLocation("brightness"), fbrightness);
					gl.glUniform1i(program.getUniformLocation("overflow"), overflow.ordinal());
					gl.glUniform1i(program.getUniformLocation("complexity"), ceiledComplexity);
					gl.glUniform1fv(program.getUniformLocation("influences[0]"), ceiledComplexity, influences, 0);
					gl.glUniformMatrix3fv(program.getUniformLocation("matrices[0]"), ceiledComplexity, false, matrices, 0);
					gl.glUniform1f(program.getUniformLocation("evolution"), fevolution);
					gl.glUniform1f(program.getUniformLocation("cycle"), fcycle);

					gl.glBegin(GL2.GL_QUADS);
					gl.glVertex2f(0, 0);
					gl.glVertex2f(w, 0);
					gl.glVertex2f(w, h);
					gl.glVertex2f(0, h);
					gl.glEnd();
				}
			});

//			gl.glActiveTexture(GL2.GL_TEXTURE1);
//			gl.glBindTexture(GL2.GL_TEXTURE_2D, 0);
			gl.glActiveTexture(GL2.GL_TEXTURE0);
			gl.glBindTexture(GL2.GL_TEXTURE_2D, 0);

			gl.glFramebufferTexture2D(GL2.GL_FRAMEBUFFER,
					GL2.GL_COLOR_ATTACHMENT0, GL2.GL_TEXTURE_RECTANGLE, 0, 0);

			IVideoBuffer result = buffer;
			buffer = null;
			return result;

		} finally {
			if (buffer != null) buffer.dispose();
		}
	}

	private IVideoBuffer doBlend(
			final GL2 gl, GLU glu, IVideoBuffer noise, IVideoBuffer input,
			final BlendMode blendMode, final double opacity) {

		VideoBounds bounds = input.getBounds();
		final int w = bounds.width;
		final int h = bounds.height;

		VideoEffectUtil.ortho2D(gl, glu, w, h);

		IVideoBuffer buffer = context.createVideoBuffer(bounds);
		try {
			gl.glFramebufferTexture2D(GL2.GL_FRAMEBUFFER,
					GL2.GL_COLOR_ATTACHMENT0, GL2.GL_TEXTURE_RECTANGLE, buffer.getTexture(), 0);
			gl.glDrawBuffer(GL2.GL_COLOR_ATTACHMENT0);

			gl.glActiveTexture(GL2.GL_TEXTURE0);
			gl.glBindTexture(GL2.GL_TEXTURE_RECTANGLE, noise.getTexture());
			gl.glActiveTexture(GL2.GL_TEXTURE1);
			gl.glBindTexture(GL2.GL_TEXTURE_RECTANGLE, input.getTexture());

			final IShaderProgram program = shaders.getProgram(FractalNoise.class, blendMode.name());
			program.useProgram(new Runnable() {
				public void run() {
					gl.glUniform1i(program.getUniformLocation("texNoise"), 0);
					if (blendMode != BlendMode.NONE) {
						gl.glUniform1i(program.getUniformLocation("texOrig"), 1);
					}
					gl.glUniform1f(program.getUniformLocation("opacity"), (float)opacity);

					gl.glBegin(GL2.GL_QUADS);
					gl.glTexCoord2f(0, 0);
					gl.glVertex2f(0, 0);
					gl.glTexCoord2f(w, 0);
					gl.glVertex2f(w, 0);
					gl.glTexCoord2f(w, h);
					gl.glVertex2f(w, h);
					gl.glTexCoord2f(0, h);
					gl.glVertex2f(0, h);
					gl.glEnd();
				}
			});

			gl.glActiveTexture(GL2.GL_TEXTURE1);
			gl.glBindTexture(GL2.GL_TEXTURE_RECTANGLE, 0);
			gl.glActiveTexture(GL2.GL_TEXTURE0);
			gl.glBindTexture(GL2.GL_TEXTURE_RECTANGLE, 0);

			gl.glFramebufferTexture2D(GL2.GL_FRAMEBUFFER,
					GL2.GL_COLOR_ATTACHMENT0, GL2.GL_TEXTURE_RECTANGLE, 0, 0);

			IVideoBuffer result = buffer;
			buffer = null;
			return result;

		} finally {
			if (buffer != null) buffer.dispose();
		}
	}

	@ShaderSource
	public static final String[] NONE = {
		"uniform sampler2DRect texNoise;",
		"uniform float opacity;",
		"",
		"void main(void)",
		"{",
		"	vec3 rgb = texture2DRect(texNoise, gl_TexCoord[0].st).rgb;",
		"	gl_FragColor = vec4(rgb*opacity, 1.0);",
		"}"
	};

	@ShaderSource(program=false)
	public static final String[] hsl2rgb = {
		"float hue2rgb(float t1, float t2, float hue)",
		"{",
		"	if (hue < 0.0) hue += 1.0;",
		"	else if (hue > 1.0) hue -= 1.0;",
		"",
		"	return (hue*6.0 < 1.0) ? t1 + (t2-t1)*6.0*hue :",
		"		   (hue*2.0 < 1.0) ? t2 :",
		"		   (hue*3.0 < 2.0) ? t1 + (t2-t1)*(2.0/3.0-hue)*6.0 :",
		"		                     t1;",
		"}",
		"",
		"vec3 hsl2rgb(vec3 hsl)",
		"{",
		"	float hue = hsl.x;",
		"	float sat = hsl.y;",
		"	float luma = hsl.z;",
		"	vec3 rgb;",
		"",
		"	if (sat == 0.0) {",
		"		rgb = vec3(luma);",
		"	} else {",
		"		float t2 = (luma < 0.5) ? luma*(1.0+sat) : luma+sat-luma*sat;",
		"		float t1 = luma*2.0 - t2;",
		"",
		"		rgb = vec3(",
		"				hue2rgb(t1, t2, hue+1.0/3.0),",
		"				hue2rgb(t1, t2, hue),",
		"				hue2rgb(t1, t2, hue-1.0/3.0));",
		"	}",
		"	return rgb;",
		"}",
	};

	private static String[] createHueOrSaturationBlendSource(boolean hue) {
		return new String[] {
				"uniform sampler2DRect texNoise;",
				"uniform sampler2DRect texOrig;",
				"uniform float opacity;",
				"",
				"const vec3 lumaVec = vec3(0.299, 0.587, 0.114);",
				"",
		  hue ? "vec3 hsl2rgb(vec3 hsl);" : "",
		  hue ? "float sat(vec3 color);" : "",
				"vec3 set_sat(vec3 color, float s);",
				"vec3 set_lum(vec3 color, float l);",
				"",
				"void main(void)",
				"{",
				"	vec2 texCoord = gl_TexCoord[0].st;",
				"	vec4 pSrc = texture2DRect(texNoise, texCoord);",
				"	vec4 pDst = texture2DRect(texOrig, texCoord);",
				"",
				"	float da = pDst.a;",
				"	float sa = pSrc.a*opacity;",
//				"	float cda = 1.0 - da;",
				"	float csa = 1.0 - sa;",
				"	float a = sa + da*csa;",
				"",
				"	vec3 pDst3 = pDst.rgb;",
				"	vec3 uDst3 = (pDst.a != 0.0) ? pDst.rgb/pDst.a : vec3(0.0);",
				"",
		  hue ? "	vec3 newDst3 = set_lum(set_sat(hsl2rgb(vec3(pSrc.r, 1.0, 0.5)), sat(uDst3)), dot(uDst3, lumaVec));"
			  : "	vec3 newDst3 = set_lum(set_sat(uDst3, pSrc.r), dot(uDst3, lumaVec));",
				"",
				"	gl_FragColor = vec4(newDst3*da*sa + pDst3*csa, a);",
				"}"
		};
	};

	@ShaderSource(attach={"ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions", "hsl2rgb"})
	public static final String[] HUE = createHueOrSaturationBlendSource(true);

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] SATURATION = createHueOrSaturationBlendSource(false);


	private static String[] createBlendSource(String name) {
		return new String[] {
								"uniform sampler2DRect texNoise;",
								"uniform sampler2DRect texOrig;",
								"uniform float opacity;",
								"",
				String.format(	"vec4 blend_%s(vec4 pDst, vec4 pSrc, float opacity);", name),
								"",
								"void main(void)",
								"{",
								"	vec2 texCoord = gl_TexCoord[0].st;",
								"	vec4 noise = texture2DRect(texNoise, texCoord);",
								"	vec4 orig = texture2DRect(texOrig, texCoord);",
				String.format(	"	gl_FragColor = blend_%s(orig, noise, opacity);", name),
								"}"
		};
	}

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] NORMAL = createBlendSource("normal");

//	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
//	public static final String[] DISSOLVE = createBlendSource("dissolve");
//
//	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
//	public static final String[] DANCING_DISSOLVE = createBlendSource("dancing_dissolve");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] DARKEN = createBlendSource("darken");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] MULTIPLY = createBlendSource("multiply");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] COLOR_BURN = createBlendSource("color_burn");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] LINEAR_BURN = createBlendSource("linear_burn");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] DARKER_COLOR = createBlendSource("darker_color");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] ADD = createBlendSource("add");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] LIGHTEN = createBlendSource("lighten");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] SCREEN = createBlendSource("screen");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] COLOR_DODGE = createBlendSource("color_dodge");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] LINEAR_DODGE = createBlendSource("linear_dodge");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] LIGHTER_COLOR = createBlendSource("lighter_color");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] OVERLAY = createBlendSource("overlay");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] SOFT_LIGHT = createBlendSource("soft_light");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] HARD_LIGHT = createBlendSource("hard_light");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] LINEAR_LIGHT = createBlendSource("linear_light");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] VIVID_LIGHT = createBlendSource("vivid_light");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] PIN_LIGHT = createBlendSource("pin_light");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] HARD_MIX = createBlendSource("hard_mix");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] DIFFERENCE = createBlendSource("difference");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] EXCLUSION = createBlendSource("exclusion");

//	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
//	public static final String[] HUE = createBlendSource("hue");
//
//	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
//	public static final String[] SATURATION = createBlendSource("saturation");
//
//	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
//	public static final String[] COLOR = createBlendSource("color");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] LUMINOSITY = createBlendSource("luminosity");

//	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
//	public static final String[] STENCIL_ALPHA = createBlendSource("stencil_alpha");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] STENCIL_LUMA = createBlendSource("stencil_luma");

//	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
//	public static final String[] SILHOUETTE_ALPHA = createBlendSource("silhouette_alpha");

	@ShaderSource(attach="ch.kuramo.javie.core.shaders.BlendModeShaders.blend_functions")
	public static final String[] SILHOUETTE_LUMA = createBlendSource("silhouette_luma");

}
