/*
 * Copyright (c) 2009,2010 Yoshikazu Kuramochi
 * All rights reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package ch.kuramo.javie.effects.stylize;

import javax.media.opengl.GL2;
import javax.media.opengl.glu.GLU;

import ch.kuramo.javie.api.IAnimatableBoolean;
import ch.kuramo.javie.api.IAnimatableInteger;
import ch.kuramo.javie.api.IShaderProgram;
import ch.kuramo.javie.api.IVideoBuffer;
import ch.kuramo.javie.api.VideoBounds;
import ch.kuramo.javie.api.annotations.Effect;
import ch.kuramo.javie.api.annotations.Property;
import ch.kuramo.javie.api.annotations.ShaderSource;
import ch.kuramo.javie.api.services.IShaderRegistry;
import ch.kuramo.javie.api.services.IVideoEffectContext;
import ch.kuramo.javie.effects.VideoEffectUtil;

import com.google.inject.Inject;

@Effect(id="ch.kuramo.javie.Mosaic", category=Effect.STYLIZE)
public class Mosaic {

	@ShaderSource
	public static final String[] MOSAIC_SHARP_COLORS = {
		"uniform sampler2DRect texture;",
		"uniform float hSize;",
		"uniform float vSize;",
		"",
		"void main(void)",
		"{",
		"	float s = (floor(gl_TexCoord[0].s/hSize)+0.5)*hSize;",
		"	float t = (floor(gl_TexCoord[0].t/vSize)+0.5)*vSize;",
		"	gl_FragColor = texture2DRect(texture, vec2(s, t));",
		"}"
	};


	private final IVideoEffectContext context;

	private final IShaderRegistry shaders;

	private final IShaderProgram sharpColorsProgram;

	@Property(value="10", min="1")
	private IAnimatableInteger horizontalBlocks;

	@Property(value="10", min="1")
	private IAnimatableInteger verticalBlocks;

	@Property
	private IAnimatableBoolean sharpColors;


	@Inject
	public Mosaic(IVideoEffectContext context, IShaderRegistry shaders) {
		this.context = context;
		this.shaders = shaders;
		this.sharpColorsProgram = shaders.getProgram(Mosaic.class, "MOSAIC_SHARP_COLORS");
	}

	public IVideoBuffer doVideoEffect() {
		int hBlocks = context.value(horizontalBlocks);
		int vBlocks = context.value(verticalBlocks);
		boolean sharpColors = context.value(this.sharpColors);

		if (sharpColors) {
			return doSharpColorsMosaic(hBlocks, vBlocks);
		} else {
			return doAverageColorsMosaic(hBlocks, vBlocks);
		}
	}

	private IVideoBuffer doSharpColorsMosaic(int hBlocks, int vBlocks) {
		IVideoBuffer input = context.doPreviousEffect();
		VideoBounds bounds = input.getBounds();
		final int w = bounds.width;
		final int h = bounds.height;

		hBlocks = Math.min(hBlocks, w);
		vBlocks = Math.min(vBlocks, h);

		if (hBlocks == w && vBlocks == h) {
			return input;
		}

		final float hSize = (float)w/hBlocks;
		final float vSize = (float)h/vBlocks;

		IVideoBuffer buffer = null;
		try {
			buffer = context.createVideoBuffer(bounds);

			final GL2 gl = context.getGL().getGL2();

			VideoEffectUtil.ortho2D(gl, context.getGLU(), w, h);

			gl.glFramebufferTexture2D(GL2.GL_FRAMEBUFFER,
					GL2.GL_COLOR_ATTACHMENT0, GL2.GL_TEXTURE_RECTANGLE, buffer.getTexture(), 0);
			gl.glDrawBuffer(GL2.GL_COLOR_ATTACHMENT0);

			gl.glActiveTexture(GL2.GL_TEXTURE0);
			gl.glBindTexture(GL2.GL_TEXTURE_RECTANGLE, input.getTexture());

			sharpColorsProgram.useProgram(new Runnable() {
				public void run() {
					gl.glUniform1i(sharpColorsProgram.getUniformLocation("texture"), 0);
					gl.glUniform1f(sharpColorsProgram.getUniformLocation("hSize"), hSize);
					gl.glUniform1f(sharpColorsProgram.getUniformLocation("vSize"), vSize);

					gl.glBegin(GL2.GL_QUADS);
					gl.glTexCoord2f(0, 0);
					gl.glVertex2f(0, 0);
					gl.glTexCoord2f(w, 0);
					gl.glVertex2f(w, 0);
					gl.glTexCoord2f(w, h);
					gl.glVertex2f(w, h);
					gl.glTexCoord2f(0, h);
					gl.glVertex2f(0, h);
					gl.glEnd();
				}
			});

			gl.glActiveTexture(GL2.GL_TEXTURE0);
			gl.glBindTexture(GL2.GL_TEXTURE_RECTANGLE, 0);

			gl.glFramebufferTexture2D(GL2.GL_FRAMEBUFFER,
					GL2.GL_COLOR_ATTACHMENT0, GL2.GL_TEXTURE_RECTANGLE, 0, 0);

			IVideoBuffer output = buffer;
			buffer = null;
			return output;

		} finally {
			input.dispose();
			if (buffer != null) buffer.dispose();
		}
	}

	private IVideoBuffer doAverageColorsMosaic(int hBlocks, int vBlocks) {
		IVideoBuffer input = context.doPreviousEffect();
		IVideoBuffer buffer = null;
		try {
			VideoBounds bounds = input.getBounds();
			int w = bounds.width;
			int h = bounds.height;

			hBlocks = Math.min(hBlocks, w);
			vBlocks = Math.min(vBlocks, h);

			if (hBlocks != w || vBlocks != h) {
				buffer = context.createVideoBuffer(new VideoBounds(hBlocks, vBlocks));
				average(input, buffer);
				scale(buffer, input);
			}

			IVideoBuffer output = input;
			input = null;
			return output;

		} finally {
			if (input != null) input.dispose();
			if (buffer != null) buffer.dispose();
		}
	}

	private void average(IVideoBuffer src, IVideoBuffer dst) {
		final GL2 gl = context.getGL().getGL2();

		IVideoBuffer src0 = src;
		IVideoBuffer tmpBuf = null;
		try {
			VideoBounds srcBounds = src.getBounds();
			VideoBounds dstBounds = dst.getBounds();
			int hBlocks = dstBounds.width;
			int vBlocks = dstBounds.height;
			int hbs = (srcBounds.width + hBlocks - 1) / hBlocks;
			int vbs = (srcBounds.height + vBlocks - 1) / vBlocks;

			// CONVOLUTIONフィルタのカーネルサイズが縦横それぞれ7までなので
			// ブロックサイズが7より大きい場合は複数回にわけて処理する
			if (hbs > 7 || vbs > 7) {
				tmpBuf = context.createVideoBuffer(new VideoBounds(
						(hbs > 7) ? hBlocks*7 : hBlocks, (vbs > 7) ? vBlocks*7 : vBlocks));
				average(src, tmpBuf);
				src = tmpBuf;

				srcBounds = src.getBounds();
				hbs = (srcBounds.width + hBlocks - 1) / hBlocks;
				vbs = (srcBounds.height + vBlocks - 1) / vBlocks;
			}

			final int srcWidth = srcBounds.width;
			final int srcHeight = srcBounds.height;
			final int dstWidth = dstBounds.width;
			final int dstHeight = dstBounds.height;

			final int ksize = hbs * vbs;
			final float[] kernel = new float[ksize];
			final float[] offset = new float[ksize*2];

			for (int j = 0; j < vbs; ++j) {
				for (int i = 0; i < hbs; ++i) {
					int k = j*hbs+i;
					kernel[k] = 1f/ksize;
					offset[k*2  ] = i - hbs/2;
					offset[k*2+1] = j - vbs/2;
				}
			}

			VideoEffectUtil.setClampToEdge(src, gl);
			VideoEffectUtil.ortho2D(gl, context.getGLU(), hBlocks, vBlocks);

			gl.glFramebufferTexture2D(GL2.GL_FRAMEBUFFER,
					GL2.GL_COLOR_ATTACHMENT0, GL2.GL_TEXTURE_RECTANGLE, dst.getTexture(), 0);
			gl.glDrawBuffer(GL2.GL_COLOR_ATTACHMENT0);

			gl.glActiveTexture(GL2.GL_TEXTURE0);
			gl.glBindTexture(GL2.GL_TEXTURE_RECTANGLE, src.getTexture());

			final IShaderProgram program = shaders.getProgram(VideoEffectUtil.class, "CONVOLUTION");
			program.useProgram(new Runnable() {
				public void run() {
					gl.glUniform1i(program.getUniformLocation("texture"), 0);
					gl.glUniform1i(program.getUniformLocation("ksize"), ksize);
					gl.glUniform1fv(program.getUniformLocation("kernel[0]"), ksize, kernel, 0);
					gl.glUniform2fv(program.getUniformLocation("offset[0]"), ksize, offset, 0);

					gl.glBegin(GL2.GL_QUADS);
					gl.glTexCoord2f(0, 0);
					gl.glVertex2f(0, 0);
					gl.glTexCoord2f(srcWidth, 0);
					gl.glVertex2f(dstWidth, 0);
					gl.glTexCoord2f(srcWidth, srcHeight);
					gl.glVertex2f(dstWidth, dstHeight);
					gl.glTexCoord2f(0, srcHeight);
					gl.glVertex2f(0, dstHeight);
					gl.glEnd();
				}
			});

			gl.glActiveTexture(GL2.GL_TEXTURE0);
			gl.glBindTexture(GL2.GL_TEXTURE_RECTANGLE, 0);

			gl.glFramebufferTexture2D(GL2.GL_FRAMEBUFFER,
					GL2.GL_COLOR_ATTACHMENT0, GL2.GL_TEXTURE_RECTANGLE, 0, 0);

		} finally {
			if (tmpBuf != null) tmpBuf.dispose();
			VideoEffectUtil.setClampToBorder(src0, gl);
		}
	}

	private void scale(IVideoBuffer src, IVideoBuffer dst) {
		GL2 gl = context.getGL().getGL2();
		GLU glu = context.getGLU();

		VideoEffectUtil.setNearest(src, gl);

		VideoBounds srcBounds = src.getBounds();
		VideoBounds dstBounds = dst.getBounds();

		VideoEffectUtil.ortho2D(gl, glu, dstBounds.width, dstBounds.height);

		gl.glFramebufferTexture2D(GL2.GL_FRAMEBUFFER,
				GL2.GL_COLOR_ATTACHMENT0, GL2.GL_TEXTURE_RECTANGLE, dst.getTexture(), 0);
		gl.glDrawBuffer(GL2.GL_COLOR_ATTACHMENT0);

		gl.glActiveTexture(GL2.GL_TEXTURE0);
		gl.glBindTexture(GL2.GL_TEXTURE_RECTANGLE, src.getTexture());
		gl.glEnable(GL2.GL_TEXTURE_RECTANGLE);

		gl.glColor4f(1, 1, 1, 1);

		gl.glBegin(GL2.GL_QUADS);
		gl.glTexCoord2f(0, 0);
		gl.glVertex2f(0, 0);
		gl.glTexCoord2f(srcBounds.width, 0);
		gl.glVertex2f(dstBounds.width, 0);
		gl.glTexCoord2f(srcBounds.width, srcBounds.height);
		gl.glVertex2f(dstBounds.width, dstBounds.height);
		gl.glTexCoord2f(0, srcBounds.height);
		gl.glVertex2f(0, dstBounds.height);
		gl.glEnd();

		gl.glActiveTexture(GL2.GL_TEXTURE0);
		gl.glBindTexture(GL2.GL_TEXTURE_RECTANGLE, 0);
		gl.glDisable(GL2.GL_TEXTURE_RECTANGLE);

		gl.glFramebufferTexture2D(GL2.GL_FRAMEBUFFER,
				GL2.GL_COLOR_ATTACHMENT0, GL2.GL_TEXTURE_RECTANGLE, 0, 0);
	}

}
